from dataclasses import dataclass


@dataclass
class CogReadError(Exception):
    message: str


@dataclass
class InvalidTiffError(CogReadError):
    ...


@dataclass
class TileNotFoundError(CogReadError):
    ...



from dataclasses import dataclass
import math

import numpy as np

from .compression import Compression
from .constants import COMPRESSIONS, INTERLEAVE, SAMPLE_DTYPES
from .filesystems import Filesystem
from .tag import Tag

@dataclass
class BaseIFD:
    next_ifd_offset: int
    tag_count: int
    _file_reader: Filesystem

@dataclass
class RequiredTags:
    BitsPerSample: Tag
    Compression: Tag
    ImageHeight: Tag
    ImageWidth: Tag
    PhotometricInterpretation: Tag
    PlanarConfiguration: Tag
    SampleFormat: Tag
    SamplesPerPixel: Tag
    TileByteCounts: Tag
    TileHeight: Tag
    TileOffsets: Tag
    TileWidth: Tag

@dataclass
class OptionalTags:
    NewSubfileType: Tag = None
    Predictor: Tag = None
    JPEGTables: Tag = None

    GeoKeyDirectoryTag: Tag = None
    ModelPixelScaleTag: Tag = None
    ModelTiepointTag: Tag = None

@dataclass
class IFD(OptionalTags, Compression, RequiredTags, BaseIFD):

    @property
    def compression(self):
        return COMPRESSIONS[self.Compression.value]

    @property
    def bands(self):
        return self.SamplesPerPixel.value

    @property
    def dtype(self):
        if self.bands == 1:
            return np.dtype(
                SAMPLE_DTYPES[(self.SampleFormat.value, self.BitsPerSample.value)]
            )
        else:
            return np.dtype(
                SAMPLE_DTYPES[(self.SampleFormat.value[0], self.BitsPerSample.value[0])]
            )

    @property
    def interleave(self):
        return "band" if self.bands == 1 else INTERLEAVE[self.PlanarConfiguration.value]

    @property
    def is_full_resolution(self):
        if not self.NewSubfileType:
            return True
        elif self.NewSubfileType.value[0] == 0:
            return False
        return True

    @property
    def is_mask(self):
        # # https://www.awaresystems.be/imaging/tiff/tifftags/newsubfiletype.html
        # # https://gdal.org/drivers/raster/gtiff.html#internal-nodata-masks
        if self.NewSubfileType:
            if self.NewSubfileType.value[2] == 1 and self.PhotometricInterpretation.value == 4 and self.compression == "deflate":
                return True
        return False


    @property
    def tile_count(self):
        return (
            math.ceil(self.ImageWidth.value / float(self.TileWidth.value)),
            math.ceil(self.ImageHeight.value / float(self.TileHeight.value)),
        )

    def __iter__(self):
        for (k, v) in self.__dict__.items():
            if k not in ("next_ifd_offset", "tag_count", "_file_reader") and v:
                yield v

    @classmethod
    async def read(cls, file_reader: Filesystem) -> "IFD":
        ifd_start = file_reader.tell()
        tag_count = await file_reader.read(2, cast_to_int=True)
        tiff_tags = {}
        # Read tags
        for idx in range(tag_count):
            tag = await Tag.read(file_reader)
            if tag:
                tiff_tags[tag.name] = tag
        file_reader.seek(ifd_start + (12 * tag_count) + 2)
        next_ifd_offset = await file_reader.read(4, cast_to_int=True)
        return cls(next_ifd_offset, tag_count, file_reader, **tiff_tags)



import abc
from dataclasses import dataclass
import struct
from typing import Optional

import imagecodecs
import numpy as np

from .filesystems import Filesystem
from .tag import Tag


@dataclass
class Compression(metaclass=abc.ABCMeta):
    _file_reader: Filesystem
    TileHeight: Tag
    TileWidth: Tag
    Predictor: Optional[Tag]
    JPEGTables: Optional[Tag]


    @property
    @abc.abstractmethod
    def bands(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def compression(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def dtype(self) -> np.dtype:
        ...

    def decompress(self, tile):
        try:
            return getattr(self, f"_{self.compression}")(tile)
        except AttributeError as e:
            raise NotImplementedError(
                f"{self.compression} is not currently supported"
            ) from e

    def decompress_mask(self, tile):
        decoded = np.frombuffer(imagecodecs.zlib_decode(tile), np.dtype('uint8'))
        mask = np.unpackbits(decoded).reshape(self.TileHeight.value, self.TileWidth.value) * 255
        return mask

    def _reshape(self, arr):
        return arr.reshape(
            self.TileHeight.value,
            self.TileWidth.value,
            self.bands,
        )

    def _unpredict(self, arr):
        # Unpredict if there is horizontal differencing
        if self.Predictor.value == 2:
            imagecodecs.delta_decode(arr, out=arr, axis=-1)

    def _jpeg(self, tile):
        jpeg_tables = self.JPEGTables
        jpeg_table_bytes = struct.pack(
            f"{self._file_reader._endian}{jpeg_tables.count}{jpeg_tables.tag_type.format}",
            *self.JPEGTables.value,
        )
        # # https://github.com/mapbox/COGDumper/tree/master/cogdumper
        if jpeg_table_bytes:
            if tile[0] == 0xFF and tile[1] == 0xD8:
                # insert tables, first removing the SOI and EOI
                tile = tile[0:2] + jpeg_table_bytes[2:-2] + tile[2:]
            else:
                raise Exception("Missing SOI marker for JPEG tile")
        decoded = imagecodecs.jpeg_decode(tile)
        return np.rollaxis(decoded, 2, 0)

    def _lzw(self, tile):
        decoded = self._reshape(np.frombuffer(imagecodecs.lzw_decode(tile), self.dtype))
        self._unpredict(decoded)
        return np.rollaxis(decoded, 2, 0)

    def _webp(self, tile):
        decoded = np.rollaxis(imagecodecs.webp_decode(tile), 2, 0)
        return decoded

    def _deflate(self, tile):
        decoded = self._reshape(np.frombuffer(imagecodecs.zlib_decode(tile), self.dtype))
        self._unpredict(decoded)
        return np.rollaxis(decoded, 2, 0)


import abc
from dataclasses import dataclass
from urllib.parse import urlsplit

import aioboto3
import aiofiles
import aiohttp

from .constants import HEADER_OFFSET

@dataclass
class Filesystem(abc.ABC):
    filepath: str

    def __post_init__(self):
        self.data: bytes = b""
        self._offset: int = 0
        self._endian: str = "<"
        self._total_bytes_requested: int = 0
        self._total_requests: int = 0

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ...

    @classmethod
    def create_from_filepath(cls, filepath: str) -> "Filesystem":
        splits = urlsplit(filepath)
        if splits.scheme in {"http", "https"}:
            return HttpFilesystem(filepath)
        elif splits.scheme == "s3":
            return S3Filesystem(filepath)
        elif (not splits.scheme and not splits.netloc):
            return LocalFilesystem(filepath)
        raise NotImplemented("Unsupported file system")

    @abc.abstractmethod
    async def range_request(self, start: int, offset: int) -> bytes:
        ...

    @abc.abstractmethod
    async def close(self) -> None:
        ...

    async def read(self, offset: int, cast_to_int: bool = False):
        if self._offset + offset > len(self.data):
            self.data += await self.range_request(len(self.data), HEADER_OFFSET)
        data = self.data[self._offset : self._offset + offset]
        self.incr(offset)
        order = "little" if self._endian == "<" else "big"
        return int.from_bytes(data, order) if cast_to_int else data

    def incr(self, offset: int) -> None:
        self._offset += offset

    def seek(self, offset: int) -> None:
        self._offset = offset

    def tell(self) -> int:
        return self._offset


@dataclass
class HttpFilesystem(Filesystem):

    async def range_request(self, start, offset):
        range_header = {"Range": f"bytes={start}-{start + offset}"}
        async with self.session.get(self.filepath, headers=range_header) as cog:
            data = await cog.content.read()
            self._total_bytes_requested += int(cog.headers["Content-Length"])
            self._total_requests += 1
        return data

    async def close(self):
        await self.session.close()

    async def __aenter__(self):
        self.session = aiohttp.ClientSession()
        return self


@dataclass
class LocalFilesystem(Filesystem):

    async def range_request(self, start, offset):
        await self.file.seek(start)
        self._total_bytes_requested += (offset - start)
        self._total_requests += 1
        return await self.file.read(offset+1)

    async def close(self):
        await self.file.close()

    async def __aenter__(self):
        self.file = await aiofiles.open(self.filepath, 'rb')
        return self


@dataclass
class S3Filesystem(Filesystem):

    async def range_request(self, start: int, offset: int) -> bytes:
        req = await self.object.get(Range=f'bytes={start}-{start+offset}')
        self._total_bytes_requested += int(req['ResponseMetadata']['HTTPHeaders']['content-length'])
        self._total_requests += 1
        data = await req['Body'].read()
        return data

    async def close(self) -> None:
        await self.resource.__aexit__('', '', '')

    async def __aenter__(self):
        splits = urlsplit(self.filepath)
        self.resource = await aioboto3.resource('s3').__aenter__()
        self.object = await self.resource.Object(splits.netloc, splits.path[1:])
        return self


from dataclasses import dataclass
import logging
import struct

from typing import Any, Tuple, Union

from .constants import TIFF_TAGS, HEADER_OFFSET

# TODO: Move this somewhere more sensible
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


@dataclass
class TagType:
    """
    Represents the type of a TIFF tag.  Also responsible for reading the tag since this is dependent on the tag's type.
    """

    format: str
    size: int


TAG_TYPES = {
    1: TagType(format="B", size=1),  # TIFFByte
    2: TagType(format="c", size=1),  # TIFFascii
    3: TagType(format="H", size=2),  # TIFFshort
    4: TagType(format="L", size=4),  # TIFFlong
    5: TagType(format="f", size=4),  # TIFFrational
    7: TagType(format="B", size=1),  # undefined
    12: TagType(format="d", size=8),  # TIFFdouble
    16: TagType(format="Q", size=8),  # TIFFlong8
}


@dataclass
class Tag:
    code: int
    name: str
    tag_type: TagType
    count: int
    length: int
    value: Union[Any, Tuple[Any]]

    def __getitem__(self, key):
        return self.value[key]

    def __len__(self):
        return self.count

    @classmethod
    async def read(cls, reader):
        # 0-2 bytes of tag are tag name
        code = await reader.read(2, cast_to_int=True)
        if code not in TIFF_TAGS:
            logger.warning(f"TIFF TAG {code} is not supported.")
            reader.incr(10)
            return None
        name = TIFF_TAGS[code]
        # 2-4 bytes are field type
        field_type = TAG_TYPES[(await reader.read(2, cast_to_int=True))]
        # 4-8 bytes are number of values
        count = await reader.read(4, cast_to_int=True)
        length = field_type.size * count
        if length <= 4:
            data = await reader.read(length)
            # We need to interpret both bits of NewSubfileType independently, even though the tiff spec says there is
            # a single value If we need to keep adding more custom logic here for specific tags we should switch to
            # something more declarative where each tag defines how to read its data.
            value = struct.unpack(f"{reader._endian}{count}{field_type.format}", data)
            reader.incr(4 - length)

            if name == "NewSubfileType":
                bit32 = '{:032b}'.format(value[0])
                value = [[int(x) for x in str(int(bit32)).zfill(3)]]

        else:
            value_offset = await reader.read(4, cast_to_int=True)
            end_of_tag = reader.tell()
            if value_offset + length > HEADER_OFFSET:
                data = await reader.range_request(value_offset, length - 1)
            else:
                reader.seek(value_offset)
                data = await reader.read(length)
            value = struct.unpack(f"{reader._endian}{count}{field_type.format}", data)
            reader.seek(end_of_tag)
        value = value[0] if count == 1 else value

        tag = Tag(
            code=code,
            name=name,
            tag_type=field_type,
            count=count,
            length=length,
            value=value,
        )
        return tag



from setuptools import setup, find_packages

with open("README.md") as f:
    desc = f.read()

setup(
    name="aiocogeo",
    description="Asynchronous cogeotiff reader",
    long_description=desc,
    long_description_content_type="text/markdown",
    version="0.0.2",
    author=u"Jeff Albrecht",
    author_email="geospatialjeff@gmail.com",
    url="https://github.com/geospatial-jeff/aiocogeo",
    license="mit",
    python_requires=">=3.7",
    classifiers=[
        'Intended Audience :: Developers',
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'License :: OSI Approved :: MIT License',
    ],
    keywords="cogeo COG",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=[
        "aioboto3",
        "aiofiles",
        "aiohttp<=3.6.2",
        "aiocache",
        "affine",
        "imagecodecs",
        "scikit-image",
        "typer",
    ],
    test_suite="tests",
    setup_requires=[
        'pytest-runner'
    ],
    entry_points={"console_scripts": ["aiocogeo=aiocogeo.scripts.cli:app"]},
    tests_require=[
        "mercantile",
        "morecantile",
        "pytest",
        "pytest-asyncio<0.11.0",
        "pytest-cov",
        "rasterio",
        "rio-tiler==2.0a4",
        "shapely",
    ]
)


import asyncio
from functools import wraps
import json as _json

import typer

from aiocogeo import COGReader

app = typer.Typer()


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def _make_bold(s, **kwargs):
    return typer.style(s, bold=True, **kwargs)


def _get_ifd_stats(ifds):
    ifd_stats = []
    for idx, ifd in enumerate(ifds):
        tile_sizes = [b / 1000 for b in ifd.TileByteCounts.value]
        mean_tile_size = round(sum(tile_sizes) / len(tile_sizes), 3)
        ifd_stats.append({
            'id': idx,
            'size': (ifd.ImageWidth.value, ifd.ImageHeight.value),
            'block_size': (ifd.TileWidth.value, ifd.TileHeight.value),
            'tile_sizes': {
                'min': min(tile_sizes),
                'max': max(tile_sizes),
                'mean': mean_tile_size
            }
        })
    return ifd_stats


def _create_ifd_table(ifds, start="\t"):
    table = (
        f"{start}{_make_bold('Id', underline=True):<20}{_make_bold('Size', underline=True):<27}"
        f"{_make_bold('BlockSize', underline=True):<26}{_make_bold('MinTileSize (KB)', underline=True):<33}"
        f"{_make_bold('MaxTileSize (KB)', underline=True):<33}{_make_bold('MeanTileSize (KB)', underline=True):<33}"
    )
    for stats in _get_ifd_stats(ifds):
        table += (
            f"\n\t\t{stats['id']:<8}"
            f"{'x'.join([str(val) for val in stats['size']]):<15}"
            f"{'x'.join([str(val) for val in stats['block_size']]):<14}"
            f"{stats['tile_sizes']['min']:<21}"
            f"{stats['tile_sizes']['max']:<21}"
            f"{stats['tile_sizes']['mean']:<30}"
        )
    return table

def _create_json_info(cog):
    profile = cog.profile

    info = {
        "file": cog.filepath,
        "profile": {
            "width": profile['width'],
            "height": profile['height'],
            "bands": profile['count'],
            "dtype": profile['dtype'],
            "crs": profile['crs'],
            "origin": (profile['transform'].c, profile['transform'].f),
            "resolution": (profile['transform'].a, profile['transform'].e),
            "bbox": cog.bounds,
            "compression": cog.ifds[0].compression,
            "internal_mask": cog.is_masked
        },
        "ifd": _get_ifd_stats(cog.ifds)
    }

    if cog.is_masked:
        info['mask_ifd'] = _get_ifd_stats(cog.mask_ifds)

    return info


@app.command(
    short_help="Read COG metadata.",
    help="Read COG profile, IFD, and mask IFD metadata.",
    no_args_is_help=True
)
@coro
async def info(
    filepath: str = typer.Argument(..., file_okay=True),
    json: bool = typer.Option(False, show_default=True, help="JSON-formatted response")
):
    sep = 25
    async with COGReader(filepath) as cog:
        if json:
            return typer.echo(_json.dumps(_create_json_info(cog), indent=1))
        profile = cog.profile
        typer.echo(
            f"""
        {_make_bold("FILE INFO:", underline=True)} {_make_bold(filepath)}

          {_make_bold("PROFILE")}
            {_make_bold("Width:"):<{sep}} {profile['width']}
            {_make_bold("Height:"):<{sep}} {profile['height']}
            {_make_bold("Bands:"):<{sep}} {profile['count']}
            {_make_bold("Dtype:"):<{sep}} {profile['dtype']}
            {_make_bold("Crs:"):<{sep}} {profile['crs']}
            {_make_bold("Origin:"):<{sep}} ({profile['transform'].c}, {profile['transform'].f})
            {_make_bold("Resolution:"):<{sep}} ({profile['transform'].a}, {profile['transform'].e})
            {_make_bold("BoundingBox:"):<{sep}} {cog.bounds}
            {_make_bold("Compression:"):<{sep}} {cog.ifds[0].compression}
            {_make_bold("Internal mask:"):<{sep}} {cog.is_masked}
        """
        )
        typer.echo(
            f"""\t  {_make_bold("IFD")}
            {_create_ifd_table(cog.ifds)}
        """
        )
        if cog.is_masked:
            typer.echo(
                f"""\t  {_make_bold("MASK IFD")}
                {_create_ifd_table(cog.mask_ifds, start="")}
            """
            )


@app.command(
    short_help="Create OGC TileMatrixSet.",
    help="Create OGC TileMatrixSet representation of the COG where each IFD is a unique tile matrix.",
    no_args_is_help=True
)
@coro
async def create_tms(filepath: str):
    async with COGReader(filepath) as cog:
        typer.echo(_json.dumps(cog.create_tile_matrix_set(), indent=1))



import asyncio
from dataclasses import dataclass, field
from functools import partial
import math
from typing import List, Optional
from urllib.parse import urljoin
import uuid

from aiocache import cached, Cache
import affine
import numpy as np
from skimage.transform import resize

from .constants import PHOTOMETRIC
from .errors import InvalidTiffError, TileNotFoundError
from .filesystems import Filesystem
from .ifd import IFD


@dataclass
class COGReader:
    filepath: str
    ifds: Optional[List[IFD]] = field(default_factory=lambda: [])
    mask_ifds: Optional[List[IFD]] = field(default_factory=lambda: [])

    _version: Optional[int] = 42
    _big_tiff: Optional[bool] = False


    async def __aenter__(self):
        async with Filesystem.create_from_filepath(self.filepath) as file_reader:
            self._file_reader = file_reader
            if (await file_reader.read(2)) == b"MM":
                file_reader._endian = ">"
            version = await file_reader.read(2, cast_to_int=True)
            if version == 42:
                first_ifd = await file_reader.read(4, cast_to_int=True)
                file_reader.seek(first_ifd)
                await self.read_header()
            elif version == 43:
                raise NotImplementedError("BigTiff is not yet supported")
            else:
                raise InvalidTiffError("Not a valid TIFF")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._file_reader.close()

    def __iter__(self):
        for ifd in self.ifds:
            yield ifd

    @property
    def profile(self):
        # TODO: Support nodata value
        return {
            "driver": "GTiff",
            "width": self.ifds[0].ImageWidth.value,
            "height": self.ifds[0].ImageHeight.value,
            "count": self.ifds[0].bands,
            "dtype": str(self.ifds[0].dtype),
            "transform": self.geotransform(),
            "blockxsize": self.ifds[0].TileWidth.value,
            "blockysize": self.ifds[0].TileHeight.value,
            "compress": self.ifds[0].compression,
            "interleave": self.ifds[0].interleave,
            "crs": f"EPSG:{self.epsg}",
            "tiled": True,
            "photometric": PHOTOMETRIC[self.ifds[0].PhotometricInterpretation.value],
        }

    @property
    def epsg(self):
        ifd = self.ifds[0]
        for idx in range(0, len(ifd.GeoKeyDirectoryTag), 4):
            # 2048 is geographic crs
            # 3072 is projected crs
            if ifd.GeoKeyDirectoryTag[idx] in (2048, 3072):
                return ifd.GeoKeyDirectoryTag[idx + 3]

    @property
    def bounds(self):
        gt = self.geotransform()
        tlx = gt.c
        tly = gt.f
        brx = tlx + (gt.a * self.ifds[0].ImageWidth.value)
        bry = tly + (gt.e * self.ifds[0].ImageHeight.value)
        return (tlx, bry, brx, tly)

    @property
    def overviews(self):
        return [2 ** (ifd + 1) for ifd in range(len(self.ifds) - 1)]

    @property
    def is_masked(self):
        return True if self.mask_ifds else False

    async def read_header(self):
        next_ifd_offset = 1
        while next_ifd_offset != 0:
            ifd = await IFD.read(self._file_reader)
            next_ifd_offset = ifd.next_ifd_offset
            self._file_reader.seek(next_ifd_offset)
            # Hack to correctly identify full resolution mask ifds in deflate compressed images
            # This assumes that the image ifd always comes before the mask ifd
            try:
                if ifd.is_mask or (ifd.is_full_resolution and ifd.ImageHeight.value == self.ifds[0].ImageHeight.value):
                    self.mask_ifds.append(ifd)
                else:
                    self.ifds.append(ifd)
            except IndexError:
                self.ifds.append(ifd)

    def geotransform(self, ovr_level: int = 0):
        # Calculate overview for source image
        gt = affine.Affine(
            self.ifds[0].ModelPixelScaleTag[0],
            0.0,
            self.ifds[0].ModelTiepointTag[3],
            0.0,
            -self.ifds[0].ModelPixelScaleTag[1],
            self.ifds[0].ModelTiepointTag[4],
        )
        # Decimate the geotransform if an overview is requested
        if ovr_level > 0:
            bounds = self.bounds
            ifd = self.ifds[ovr_level]
            gt = affine.Affine.translation(bounds[0], bounds[3]) * affine.Affine.scale(
                (bounds[2] - bounds[0]) / ifd.ImageWidth.value,
                (bounds[1] - bounds[3]) / ifd.ImageHeight.value,
            )
        return gt

    def _get_overview_level(self, bounds, width, height):
        """
        https://github.com/cogeotiff/rio-tiler/blob/v2/rio_tiler/utils.py#L79-L135
        """
        src_res = self.geotransform().a
        target_gt = affine.Affine.translation(
            bounds[0], bounds[3]
        ) * affine.Affine.scale(
            (bounds[2] - bounds[0]) / width, (bounds[1] - bounds[3]) / height
        )
        target_res = target_gt.a

        ovr_level = 0
        if target_res > src_res:
            # Decimated resolution at each overview
            overviews = [src_res * decim for decim in self.overviews]
            for ovr_level in range(ovr_level, len(overviews) - 1):
                ovr_res = src_res if ovr_level == 0 else overviews[ovr_level]
                if (ovr_res < target_res) and (overviews[ovr_level + 1] > target_res):
                    break
                if abs(ovr_res - target_res) < 1e-1:
                    break
            else:
                ovr_level = len(overviews) - 1

        return ovr_level


    @cached(
        cache=Cache.MEMORY,
        # Cache key comes from filepath and x/y/z coordinates of image tile
        key_builder=lambda fn,*args,**kwargs: f"{args[0].filepath}-{args[1]}-{args[2]}-{args[3]}"
    )
    async def get_tile(self, x: int, y: int, z: int) -> bytes:
        """
        https://github.com/mapbox/COGDumper/blob/master/cogdumper/cog_tiles.py#L337-L365
        """
        futures = []
        if z > len(self.ifds):
            raise TileNotFoundError(f"Overview {z} does not exist.")
        ifd = self.ifds[z]
        idx = (y * ifd.tile_count[0]) + x
        if idx > len(ifd.TileOffsets):
            raise TileNotFoundError(f"Tile {x} {y} {z} does not exist")
        offset = ifd.TileOffsets[idx]
        byte_count = ifd.TileByteCounts[idx] - 1
        tile_task = asyncio.create_task(
            self._file_reader.range_request(offset, byte_count)
        )
        futures.append(tile_task)

        if self.is_masked:
            mask_ifd = self.mask_ifds[z]
            offset = mask_ifd.TileOffsets[idx]
            byte_count = mask_ifd.TileByteCounts[idx] - 1
            mask_task = asyncio.create_task(
                self._file_reader.range_request(offset, byte_count)
            )
            futures.append(mask_task)

        img_bytes = await asyncio.gather(*futures)

        decoded = ifd.decompress(img_bytes[0])
        if self.is_masked:
            mask = mask_ifd.decompress_mask(img_bytes[1])
            decoded = np.ma.masked_where(np.broadcast_to(mask, decoded.shape)==0, decoded)

        return decoded

    def _calculate_image_tiles(self, bounds, ovr_level):
        geotransform = self.geotransform(ovr_level)
        invgt = ~geotransform
        tile_width = self.ifds[ovr_level].TileWidth.value
        tile_height = self.ifds[ovr_level].TileHeight.value

        # Project request bounds to pixel coordinates relative to geotransform of the overview
        tlx, tly = invgt * (bounds[0], bounds[3])
        brx, bry = invgt * (bounds[2], bounds[1])

        # Calculate tiles
        xmin = math.floor((tlx + 1e-6) / tile_width)
        xmax = math.floor((brx + 1e-6) / tile_width)
        ymax = math.floor((bry + 1e-6) / tile_height)
        ymin = math.floor((tly + 1e-6) / tile_height)

        tile_bounds = (
            xmin * tile_width,
            ymin * tile_height,
            (xmax + 1) * tile_width,
            (ymax + 1) * tile_height,
        )

        # Create geotransform for the fused image
        _tlx, _tly = geotransform * (tile_bounds[0], tile_bounds[1])
        fused_gt = affine.Affine(
            geotransform.a,
            geotransform.b,
            _tlx,
            geotransform.d,
            geotransform.e,
            _tly
        )
        inv_fused_gt = ~fused_gt
        xorigin, yorigin = [round(v) for v in inv_fused_gt * (bounds[0], bounds[3])]
        return {
            "tlx": xorigin,
            "tly": yorigin,
            "width": round(brx - tlx),
            "height": round(bry - tly),
            "tile_ranges": (xmin, ymin, xmax, ymax),
        }

    @staticmethod
    def stitch_image_tile(fut, fused_arr, idx, idy, tile_width, tile_height):
        img_arr = fut.result()
        fused_arr[
            :,
            idy * tile_height : (idy + 1) * tile_height,
            idx * tile_width : (idx + 1) * tile_width
        ] = img_arr

    async def read(self, bounds, shape):
        # Determine which tiles intersect the request bounds
        ovr_level = self._get_overview_level(bounds, shape[1], shape[0])
        ifd = self.ifds[ovr_level]
        tile_height = ifd.TileHeight.value
        tile_width = ifd.TileWidth.value
        img_tiles = self._calculate_image_tiles(bounds, ovr_level)
        xmin, ymin, xmax, ymax = img_tiles["tile_ranges"]

        # Request those tiles
        tile_tasks = []
        fused = np.zeros(
            (
                ifd.bands,
                (ymax + 1 - ymin) * tile_height,
                (xmax + 1 - xmin) * tile_width,
            )
        ).astype(ifd.dtype)
        for idx, xtile in enumerate(range(xmin, xmax + 1)):
            for idy, ytile in enumerate(range(ymin, ymax + 1)):
                get_tile_task = asyncio.create_task(
                    self.get_tile(xtile, ytile, ovr_level)
                )
                get_tile_task.add_done_callback(
                    partial(
                        self.stitch_image_tile,
                        fused_arr=fused,
                        idx=idx,
                        idy=idy,
                        tile_width=tile_width,
                        tile_height=tile_height,
                    )
                )
                tile_tasks.append(get_tile_task)
        await asyncio.gather(*tile_tasks)

        # Clip to request bounds
        clipped = fused[
            :,
            img_tiles['tly']: img_tiles['tly'] + img_tiles['height'],
            img_tiles['tlx']: img_tiles['tlx'] + img_tiles['width']
        ]

        # Resample to match request size
        resized = resize(
            clipped, output_shape=(ifd.bands, shape[0], shape[1]), preserve_range=True, anti_aliasing=True
        ).astype(ifd.dtype)

        return resized

    def create_tile_matrix_set(self, identifier: str = None):
        matrices = []
        for idx, ifd in enumerate(self.ifds):
            gt = self.geotransform(idx)
            matrix = {
                "identifier": str(len(self.ifds) - idx - 1),
                "topLeftCorner": [gt.c, gt.f],
                "tileWidth": ifd.TileWidth.value,
                "tileHeight": ifd.TileHeight.value,
                "matrixWidth": ifd.tile_count[0],
                "matrixHeight": ifd.tile_count[1],
                "scaleDenominator": gt.a / 0.28e-3,
            }
            matrices.append(matrix)

        tms = {
            "title": f"Tile matrix for {self.filepath}",
            "identifier": identifier or str(uuid.uuid4()),
            "supportedCRS": urljoin(f"http://www.opengis.net", f"/def/crs/EPSG/0/{self.epsg}"),
            "tileMatrix": list(reversed(matrices))
        }
        return tms
